#pragma once

#include"Config.h"

class Reader
{
private:
	HANDLE input;
	HANDLE output;

	void Write(const tstring &str)
	{
		DWORD length = lstrlen(str.c_str());
		WriteConsole(this->output, str.c_str(), length, &length, NULL);
	}

	template<typename Pre1,typename Pre2>
	tstring PreRead(const Pre1 &loop_func, const Pre2 &body)
	{
		INPUT_RECORD input_record;
		DWORD buff;

		tstring ret;

		if (ReadConsoleInput(this->input, &input_record, 1, &buff))
		{
			auto KeyEvent = [&input_record]()
			{
				auto key_event = input_record.EventType == KEY_EVENT;
				auto key_down = input_record.Event.KeyEvent.bKeyDown;
				auto ret = key_event&&key_down;

				return ret;
			};

			if (KeyEvent())
			{
				do
				{
					bool is_shift = false;
					if (KeyEvent())
					{
						auto temp = body(&input_record);
						if (input_record.Event.KeyEvent.wVirtualKeyCode == VK_BACK)
						{
							if (!ret.empty())
							{
								ret.pop_back();
							}
						}
						else
						{
							ret += temp;
						}
					}
					ReadConsoleInput(this->input, &input_record, 1, &buff);
				} while (loop_func(input_record));
			}
		}

		FlushConsoleInputBuffer(this->input);

		return ret;
	}

public:
	Reader(const HANDLE &input, const HANDLE &output)
		:input(input), output(output) {}

	tstring Line(const bool &visible=true)
	{
			auto loop_func = [](const INPUT_RECORD &input_record)
			{
				auto key_enter = input_record.Event.KeyEvent.wVirtualKeyCode == VK_RETURN;
				return !key_enter;
			};
		
			TCHAR last_input;
			auto body = [this, &visible,&last_input](INPUT_RECORD *input_record)
			{
				tstring ret;

				auto word = input_record->Event.KeyEvent.uChar.UnicodeChar;
				auto is_print = iswprint(word);

				if (is_print)
				{
					ret = word;
				}

				if (visible)
				{
					if (input_record->Event.KeyEvent.wVirtualKeyCode == VK_BACK)
					{
						CONSOLE_SCREEN_BUFFER_INFOEX info{ sizeof(CONSOLE_SCREEN_BUFFER_INFOEX) };
						GetConsoleScreenBufferInfoEx(this->output, &info);

						DWORD written;
						
						TCHAR filler=' ';
						
						int shifter = isleadbyte(last_input);
						WORD result;
						GetStringTypeEx(LOCALE_SYSTEM_DEFAULT, CT_CTYPE3, &last_input, 1, &result);

						if (result&C3_FULLWIDTH)
						{
							filler = '@';
							shifter = 2;
						}
						info.dwCursorPosition.X -= shifter;
						
						FillConsoleOutputCharacter(
							this->output,
							filler,
							shifter,
							info.dwCursorPosition,
							&written
						);
						SetConsoleCursorPosition(this->output, info.dwCursorPosition);
					}
					this->Write(ret);
				}

				if (is_print)
				{
					last_input = word;
				}

				return ret;
			};

		auto ret = this->PreRead(loop_func,body);

		return ret;
	}

	TCHAR Word(const bool &visible = false)
	{
		auto loop_func = [](const INPUT_RECORD &input_handle)
		{
			return false;
		};

		auto body = [this, &visible](INPUT_RECORD *input_record)
		{
			tstring ret;

			ret =tolower(input_record->Event.KeyEvent.wVirtualKeyCode);
			if (visible)
			{
				this->Write(ret);
			}

			return ret;
		};

		auto str = this->PreRead(loop_func,body);
		
		TCHAR ret;
		if (str.empty())
		{
			ret = 0;
		}
		else
		{
			ret = str.front();
		}
		return ret;
	}

	void SetCursorPos(const COORD &coord)
	{
		SetConsoleCursorPosition(this->output, coord);
	}

	tstring PosLine(const COORD &coord)
	{
		this->SetCursorPos(coord);
		return this->Line();
	}
};