#include"ConsoleConfig/ConsoleConfig.h"

#include<vector>
#include<random>
#include<memory>
#include<algorithm>

using namespace std;

#define TEST

struct Coordinate
{
	double x, y;
	Coordinate operator+(const Coordinate &coord)const
	{
		Coordinate ret(*this);
		ret.x += coord.x;
		ret.y += coord.y;

		return ret;
	}

	Coordinate& operator+=(const Coordinate &coord)
	{
		*this = *this + coord;

		return *this;
	}

	operator tstring()const
	{
		auto x = to_tstring(this->x);
		auto y = to_tstring(this->y);

		tstring ret = TEXT("x:") + x + TEXT("\t") + TEXT("y:") + y;

		return ret;
	}
};

enum struct Direction
{
	LEFT,RIGHT,
};

class SlopeMap
{
	struct MapInfo
	{
		double dist, angle;
		MapInfo()
		{
			auto deg2angle = [](const double &x)
			{
				auto pi = 3.14;
				auto ref_deg = 180.;

				auto ret = pi*x / ref_deg;
				return ret;
			};

			random_device rnd;
			mt19937 mt(rnd());

			auto min_dist = 10.;
			auto max_dist = 20.;
			uniform_real_distribution<> dist(min_dist, max_dist);

			auto min_angle = 0.;
			auto max_angle = deg2angle(80);
			uniform_real_distribution<> angle(min_angle, max_angle);

			this->dist = dist(mt);

#ifdef TEST
			this->angle = deg2angle(45);
#else
			this->angle = angle(mt);
#endif
		}
	};

	vector<MapInfo> map_info;

	SlopeMap()
	{
		random_device rnd;
		mt19937 mt(rnd());

		auto min_slope = 3;
		auto max_slope = 10;
		uniform_int_distribution<> slope(min_slope, max_slope);

		this->map_info.resize(slope(mt));
	}

public:
	static unique_ptr<SlopeMap>& Generate()
	{
		struct Impl:public SlopeMap{};
		static unique_ptr<SlopeMap> obj = make_unique<Impl>();

		return obj;
	}

	MapInfo GetInfo(const double &x)
	{
		auto sum = 0.;
		auto func = [&sum,&x](const MapInfo &info)
		{
			sum += info.dist;
			auto ret= x < sum;
			return ret;
		};

		auto itr=find_if(begin(this->map_info), end(this->map_info), func);

		return *itr;
	}

	double GetAngle(const double &x)
	{
		auto info = this->GetInfo(x);
		auto ret = info.angle;

		return ret;
	}

	Coordinate GetDelta(const double &x, const Direction &dir,const double &delta=1.)
	{
		auto angle = this->GetAngle(x);

		auto factor = dir == Direction::RIGHT ? delta : -delta;
		auto dx = factor*cos(angle);
		auto dy = factor*sin(angle);
		Coordinate ret{ dx,dy };

		return ret;
	}


};

void Move(Coordinate *coord)
{
	auto input=ConsoleConfig::Generate()->WordRead();
	
	auto is_left = input == TEXT('a');
	auto is_right = input == TEXT('d');

	Direction dir=is_left?Direction::LEFT:Direction::RIGHT;

	if (is_left||is_right)
	{
		auto delta = SlopeMap::Generate()->GetDelta(coord->x, dir);
		auto temp = *coord + delta;

		auto min_x = 0;
		auto max_x = 100;

		auto too_small = temp.x < min_x;
		auto too_big = temp.x > max_x;

		if (!too_small && !too_big)
		{
			*coord = temp;
		}
	}
}



void Disp(const Coordinate &coord)
{
	Display locate{ COORD{5,10}, coord };
	ConsoleConfig::Generate()->PosWrite(locate);

	auto _angle = SlopeMap::Generate()->GetAngle(coord.x);
	Display angle{ COORD{5,12},TEXT("angle:") + to_tstring(_angle) };
	ConsoleConfig::Generate()->PosWrite(angle);
}

int main()
{
	Coordinate coord{ 0,0 };

	while (1)
	{
		Move(&coord);
		Disp(coord);
		ConsoleConfig::Generate()->Update();	
	}
}